# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re


class Topic(object):
    """A topic is a word or a small phrase to identify proper build steps.

    Class arguments:
        TOPIC         (str): Project topic to lookup.
        TOPIC_MXLIMIT (int): Max. string limit for topic lookup.
        TOPIC_PATTERN (exp): Regular expression to match topci pattern.
        PATTERN     (RegEx): Compiled regex.
        __topics     (list): Private list of topic instances.

    Args:
        order   (int): Order of appearance in build steps.
        title   (str): Topic name as found by pattern matching.
        steps  (list): List of instructions.
    """

    TOPIC = r""
    TOPIC_MXLIMIT = 24
    TOPIC_PATTERN = None
    PATTERN = None

    __topics = []

    def __init__(self, order=None, title=None, steps=None):
        self.order = order
        self.title = title
        self.steps = steps
        self.__topics.append(self)

    def set_position(self, value):
        """Topic position setter.

        Args:
            value (int): Topic position order.

        Raises:
            TypeError: If value is not an unsigned interger.
        """

        if not isinstance(value, int) or value < 0:
            raise TypeError("Position must be an unsigned integer")
        self.order = value

    def get_position(self):
        """Topic position getter.

        Returns:
            int: Topic order.
        """

        return self.order

    def set_title(self, value):
        """Topic setter for title.

        Args:
            value (str): Topic title.

        Raises:
            TypeError: If value is not a string.
        """

        if not isinstance(value, (str, unicode)):
            raise TypeError("Unsupported title")
        self.title = value

    def get_title(self):
        """Topic getter for title.

        Returns:
            str: Title of topic.
        """

        return self.title

    def has_step(self, step):
        """Lookup step string in steps list.

        Args:
            step (str): Step string to lookup.

        Returns:
            bool: True if step exists, otherwise False.
        """

        for s in self.steps:
            if s.get_step().lower() == step.lower():
                return True
        return False

    def add_step(self, step):
        """Appends step to topic instance.

        Args:
            step (str): Step string managed by current topic.
        """

        if self.steps is None:
            self.steps = []
        self.steps.append(step)

    def set_steps(self, value):
        """Topic setter for instructions list.

        Args:
            value (list): Instructions list.

        Raises:
            TypeError: If value is not a valid list.
        """

        if not isinstance(value, list):
            raise TypeError("Unsupported instructions list")
        self.steps = value

    def get_steps(self):
        """Topic getter for instructions list.

        Returns:
            list: List of instructions.
        """

        return self.steps

    @classmethod
    def get_all_topics(cls):
        """Return stored topic instances.

        Returns:
            list: List of stored topics.
        """

        return cls.__topics

    @classmethod
    def set_project_topic_pattern(cls, topic_pattern):
        """Change project topic pattern.

        A project topic pattern is a regular expression used to find and
        extract topics from a build file. By default it matches all lines
        containing the word "build" in the first 10 characters.

        Args:
            pattern (str): New topic pattern.

        Raises:
            ValueError: If project topic is not a string.
            Exception:  If project topic exceeds MAX_TOPIC_LIMIT.
        """

        if not isinstance(topic_pattern, (str, unicode)):
            raise ValueError("Project topic pattern must be string")
        cls.TOPIC_PATTERN = topic_pattern

    @classmethod
    def set_project_topic(cls, project_topic):
        """Change project topic.

        A project topic is a word or a phrase to lookup when scanning the
        entire build file. Default is "build instructions".

        Args:
            project_topic (str): New project topic.

        Raises:
            ValueError: If project topic is not a string.
            Exception:  If project topic exceeds MAX_TOPIC_LIMIT.
        """

        if not isinstance(project_topic, (str, unicode)):
            raise ValueError("Project topic must be string")
        if len(project_topic) > cls.TOPIC_MXLIMIT:
            raise Exception("Project topic exceeds characters limit")
        cls.TOPIC = project_topic

    @classmethod
    def prepare(cls, default=r"\w+"):
        """Prepare filter.

        Args:
            default (str): A default pattern to fallback.

        Returns:
            Topic: class instance.
        """

        if cls.TOPIC_PATTERN is None:
            cls.TOPIC_PATTERN = default
        cls.PATTERN = re.compile(cls.TOPIC_PATTERN, re.I)
        return cls

    def __repr__(self):
        return unicode(self.__dict__)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from timeit import default_timer
from os import getpid

from buildok.converter import Converter
from buildok.matcher import Matcher
from buildok.reader import Reader
from buildok.report import Report

from buildok.readers.read_me import ReadmeReader
from buildok.structures.topic import Topic
from buildok.statements.invoke import InvokeTopic
from buildok.util.log import Log


PID = getpid()

RUN_INSTRUCTIONS = """
    Buildok is ready! Choose a topic to run, either by it's ID or by name (can
    be partial match as well). During the runtime, SIGINT is ignored. Typing
    an ID or a name that does not exist will restart the promter. Leaving the
    field blank will end session.
"""

WARNING_UNSAFE_SHELL = u"""\033[91m
    Detected --unsafe-shell option! Use this option on your own risk only if
    you really know what you're doing! Please check documentation if you're not
    sure what --unsafe-shell is, or better yet, remove it and update your
    guide.\033[0m
"""


class Script(object):
    """Automated script wrapper.

    Args:
        args         (list): List of shell arguments.
        topic         (str): Topic as a text string.
        steps        (list): List of topic steps.
        last_step     (int): Last step index.
        guide_topics (list): List of guide topics.
        convert      (bool): Convertion flag.
    """

    def __init__(self, args):
        self.args = args
        self.topic = None
        self.steps = None
        self.last_step = 0
        self.guide_topics = None
        self.convert = False
        self.fake_run = False
        Log.info("Initializing...")

    def setup(self):
        """Setup project path, topic and topic pattern, convertion type.
        """

        # Look for a specific topic
        if self.args.topic is not None:
            Topic.set_project_topic(self.args.topic)
            Log.info("Topic changed to: %s" % self.args.topic)

        # Set topic pattern
        if self.args.topic_pattern is not None:
            Topic.set_project_topic_pattern(self.args.topic_pattern)
            Log.info("Topic pattern changed to: %s" % self.args.topic_pattern)

        # Change project path from current working directory to choosen path
        if self.args.guide is not None:
            Reader.set_project_path(self.args.guide)
            Log.info("Project guide location: %s" % self.args.guide)

        # Convert build steps to script before exit
        if self.args.convert is not None:
            Converter.prepare(self.args.convert)
            Log.info("Topic convertion set to: %s" % self.args.convert)
            self.convert = True

        # Toggle fake run
        self.fake_run = self.args.fake_run

    def run(self, ignore_fails=False):
        """Run all steps for the current selected topic.
        """

        if self.fake_run:
            Log.info("Topic '%s' faked run" % self.topic.get_title())
            Report.set_status("FAKED")
        else:
            self.launch_topic(ignore_fails)

        if self.convert:
            self.launch_convertion()

    def launch_convertion(self):
        """Convert build steps to target convertion.
        """

        Log.debug("Converting...")
        Converter.check() and Converter.save(self.steps)
        Log.debug("Conversion done!")

    def launch_topic(self, ignore_fails=False, failed=False):
        """Launch topic and run all steps.
        """

        total_steps = len(self.topic.get_steps())
        Log.info("Preparing to run %d steps from topic..." % total_steps)

        # Loop steps and run each one
        start_time = default_timer()
        Log.debug("Setting start time: %s" % start_time)

        success_log = u"\033[92m\u2713 (Success)\033[0m \033[93m%s\033[0m"
        failed_log = u"\033[91m? (Failed)\033[0m \033[95m%s\033[0m"
        error_log = u"\033[91m? (Error) %s\033[0m"

        for self.last_step, step in enumerate(self.steps):
            step_count, step_desc = self.last_step + 1, step.get_description()
            Log.debug("Preparing step (%d) %s" % (step_count, step_desc))
            Log.info(u"Running \033[93m%s\033[0m ..." % step.get_step())
            try:
                success, output = step.run()
                if success:
                    Log.info(success_log % output)
                else:
                    Log.info(failed_log % output)
                    if not ignore_fails:
                        failed = True
                        Report.set_error(output)
                        break
                Report.inc_step(1)
            except KeyboardInterrupt:
                Log.info("Stopping current step...")
                Log.warn("Interrupting may lead to unexpected results")
                Log.warn("Stop master process at your own risk (PID %s)" % PID)
            except Exception as e:
                failed = True
                Report.set_error(e)
                Log.info(error_log % str(e))
                break

        stop_time = default_timer()
        Log.debug("Set stop time: %s" % stop_time)
        Log.debug("Runtime %ss" % (stop_time - start_time))

        # Save runtime
        Report.set_runtime(stop_time - start_time)

        # Wrap up...
        Log.debug("Done running steps...")
        title = self.topic.get_title()
        if failed:
            Log.info("An error occured while topic '%s' was running" % title)
            Report.set_status("Failed")
        else:
            Log.info("Topic '%s' has ran all steps with no errors" % title)
            Report.set_status("OK")
        Log.debug("Closing...")

    def parse(self):
        """Parse guide and extract topics.

        User prompter asks to choose a topic from the selection in order to
        know what steps are queued.
        """

        rr = ReadmeReader(validate=True)
        rr.read()
        rr.parse()
        Log.info("Parsing guide...")

        # Preview scanned guide
        if self.args.preview:
            Log.info("Loading guide preview...")
            rr.preview(150)

        # Scan guide for topics
        if self.args.topic is None:
            guide = rr.get_guide()
        else:
            guide = rr.get_guide_by_topic()
        if guide is None:
            Log.fatal("Guide has no such topic")

        # Prepare guide topics
        self.guide_topics = guide.get_topics()

        # Pair each step with appropriate statement
        if self.args.strict:
            if not all([self.pair_steps(t) for t in self.guide_topics]):
                Report.set_status("Halted")
                Report.set_error("Unsupported steps")
                Log.fatal("Cannot continue because of unsupported steps")
        else:
            for t in self.guide_topics:
                for step in t.get_steps():
                    pair = Matcher.pair_one(step)
                    if not pair:
                        self.guide_topics.remove(t)

        # Scan topics
        topics = [t.get_title() for t in guide.get_topics()]
        topics_len = len(topics)

        # Handle no topics
        if topics_len == 0:
            Log.fatal("No topics found")
        else:
            Log.info("Guide has %d topics" % topics_len)

        # Handle topics
        Log.info("Found the following topics...")
        print("")
        for pos, name in enumerate(topics):
            print(("%" + str(len(str(topics_len))+5)+"d) %s") % (pos+1, name))
        print(RUN_INSTRUCTIONS)

        # Display warnings
        if self.args.unsafe_shell:
            print(WARNING_UNSAFE_SHELL)

        # Save topic and topic's steps
        self.topic = self.get_user_input()
        self.steps = self.topic.get_steps()

        # Confirm topic
        print("")
        Log.info(u"Matching topic \033[92m%s\033[0m" % self.topic.get_title())

        # Patch topic imports
        self.patch_topic_invoke(self.steps)

        # Prepare report
        Report.set_total_steps(len(self.steps))
        Report.set_topic(self.topic.get_title())
        return self

    def pair_steps(topic):
        """Pair topic steps.
        """

        return Matcher.pair_all(topic.get_steps())

    def patch_topic_invoke(self, steps):
        """Scan steps for topic invoking.

        Replace invoker step with topic's steps.
        """

        for i, step in enumerate(steps):

            # Skip if step is not an invoker
            if step.action is not InvokeTopic:
                continue

            # Find invoked topic
            topic = step.get_kwarguments().get("topic", "")
            topic_steps = self.get_steps_by_topic(topic)
            self.patch_topic_invoke(topic_steps)
            self.steps[i:i+1] = topic_steps

    def get_steps_by_topic(self, topic_name):
        """Retrieve steps by topic name.

        Returns an empty list if nothing is found.
        """

        for topic in Topic.get_all_topics():
            if topic_name.lower() == topic.get_title().lower():
                return topic.get_steps()
        return []

    def get_user_input(self, attempt=0):
        """Prompt user to choose a topic from a list.

        Exists if user leaves blank input or hits C^, otherwise continues
        until it receives a valid topic.
        """

        # Prompt for input
        try:
            if attempt > 0:
                print("")  # spacing only
            user_input = raw_input("    >>> choose topic: ")
            attempt += 1

        # Exit on C^
        except KeyboardInterrupt:
            print("")  # append newline after ^C
            Report.set_status("Exit")
            raise Exception

        # Exit on blank input
        if len(user_input) == 0:
            Report.set_status("Exit")
            raise Exception

        # Validate index or start over
        try:
            pos = int(user_input)
            if pos <= 0 or pos > len(self.guide_topics):
                raise IndexError
            return self.guide_topics[pos-1]
        except IndexError:
            print("    <<< topic #%d not found" % pos)
            return self.get_user_input(attempt)
        except Exception:
            pass

        # Validate full or partial topic name or start over
        try:
            topic = unicode(user_input)
            partial_match = None
            drop_partial = False
            for each in self.guide_topics:
                if each.get_title().strip() == topic.strip():
                    return each
                prefixed = each.get_title().startswith(topic)
                if partial_match is not None and prefixed and not drop_partial:
                    drop_partial = True
                if partial_match is None and prefixed:
                    partial_match = each
            if not drop_partial and partial_match is not None:
                return partial_match
            print("    <<< topic \"%s\" not found" % topic)
        except Exception:
            pass

        # No valid topic found, start over
        return self.get_user_input(attempt)

    def print_report(self):
        """Dumps a report of the script outcome.
        """

        Log.info("Preparing to print report...")
        Report.output()



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import path as path
from glob import glob
from shutil import copy
from distutils.dir_util import copy_tree

from buildok.action import Action


class Copy(Action):
    r"""Copy files from a given source to a given destination.

    Args:
        src (str): Source of files.
        dst (str): Target destination of files.

    Retuns:
        str: Status of copy.

    Raises:
        OSError: If an invalid `src` or `dst` is provided.

    Accepted statements:
        ^copy from `(?P<src>.+)` to `(?P<dst>.+)`$
        ^copy `(?P<src>.+)` files to `(?P<dst>.+)`$
        ^copy `(?P<src>.+)` to `(?P<dst>.+)`$

    Sample input:
        - Run `touch /tmp/buildok_test_copy.txt`.
        - Copy `/tmp/buildok_test_copy.txt` to `/tmp/buildok_test_copy2.txt`.

    Expected:
        Copied => 1 file(s) 0 dir(s)
    """

    def run(self, src=None, dst=None, *args, **kwargs):
        files, folders = 0, 0
        for item in glob(src):
            if path.isfile(item):
                copy(item, dst)
                files += 1
            elif path.isdir(item):
                copy_tree(item, dst)
                folders += 1
        self.success("Copied => %d file(s) %d dir(s)" % (files, folders))

    @classmethod
    def convert_shell(cls, src=None, dst=None, *args, **kwargs):
        if src is None and dst is None:
            return "echo invalid copy command"
        elif src is None:
            src = "."
        elif dst is None:
            dst = "."
        flags = kwargs.get("flags", "")
        if path.isdir(src):
            flags = " -R"
        return "cp%s %s %s" % (flags, src, dst)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from re import compile, IGNORECASE, UNICODE


class Instruction(object):
    """Blueprint of an instruction.

    Class arguments:
        PATTERN  (RegEx): Regular expression of an instruction.
        RunType (object): Class object of steps punctuation.

    Args:
        order       (int): Runtime order as found in topic.
        step        (str): Instruction step as text.
        punct       (str): Punctuation step as text.
        payload     (str): Instruction payload as input.
        action (callable): Action handler from installed statements.
        args      (tuple): Action arguments after applying expression to step.
        kwargs     (dict): Action keyword arguments after applying expression.
        status     (bool): Instruction status after action handler run.
        output      (str): Instruction output after action handler run.

    Raises:
        TypeError: If invalid datatype is provided.
    """

    PATTERN = compile(
        r"^(?:\-|\d+\))\s+(?P<step>.+)(?<=[^\.\:\;])(?P<punct>[\.\:\;]{1})$",
        IGNORECASE | UNICODE
    )

    class RunType(object):
        """Statement action punctuation.

        Used to determine action type for a statement.

        Possible values:
          . = Run statement.
          ; = Run statement and continue only if succeeds.
          : = Run statement with content as input params.
        """

        END = "."
        AND = ";"
        ARGS = ":"  # has payload

    def __init__(self, order=None, step=None, punct=None):
        self.order = order
        self.step = step
        self.punct = punct
        self.action = None
        self.args, self.kwargs = None, None
        self.payload = None
        self.description = None
        self.status = None
        self.output = None

    def get_status(self):
        """Instruction status getter.

        Returns:
            bool: Instruction run status.
        """

        return self.status

    def get_output(self):
        """Instruction output getter.

        Returns:
            str: Instruction run output.
        """

        return self.output

    def set_payload(self, payload):
        """Instruction payload setter.

        Has support for sequences-based datatypes.

        Args:
            payload (mixt): Instruction payload.

        Raises:
            TypeError: If unsupported data is provided.
        """

        if isinstance(payload, (tuple, list, set, frozenset)):
            self.payload = "\n".join(payload)
        elif isinstance(payload, (str, unicode)):
            self.payload = payload
        else:
            raise TypeError("Unsupported payload provided")

    def get_payload(self):
        """Instruction payload getter.

        Returns:
            str: Instruction payload.
        """

        return self.payload

    def set_position(self, value):
        """Instruction position setter.

        Args:
            value (int): Instruction position order.

        Raises:
            TypeError: If value is not an unsigned interger.
        """

        if not isinstance(value, int) or value < 0:
            raise TypeError("Position must be an unsigned integer")
        self.order = value

    def get_position(self):
        """Instruction position getter.

        Returns:
            int: Instruction order.
        """

        return self.order

    def set_step(self, value):
        """Instruction step setter.

        Args:
            value (str): Instruction raw step as text.

        Raises:
            TypeError: If value is not a non-empty string.
        """

        if not isinstance(value, (str, unicode)) or len(value) == 0:
            raise TypeError("Step must be non-empty string")
        self.step = value

    def get_step(self):
        """Instruction text step getter.

        Returns:
            str: Instruction text step string.
        """

        return self.step

    def set_description(self, value):
        """Instruction description setter.

        Args:
            value (str): Instruction description as text.

        Raises:
            TypeError: If value is not a non-empty string.
        """

        if not isinstance(value, (str, unicode)) or len(value) == 0:
            raise TypeError("Description must be non-empty string")
        self.description = value

    def get_description(self):
        """Instruction description getter.

        Returns:
            str: Instruction description string.
        """

        return self.description

    def set_statement(self, value):
        """Instruction action statement setter.

        Args:
            value (callable): Instruction step handler.

        Raises:
            TypeError: If value is not callable.
        """

        if not callable(value):
            raise TypeError("Step handler must be callable")
        self.action = value

    def get_statement(self):
        """Instruction action statement getter.

        Returns:
            callable: Instruction step handler.
        """

        return self.action

    def set_arguments(self, value):
        """Instruction action arguments setter.

        Args:
            value (tuple): Instruction step arguments.

        Raises:
            TypeError: If value is not tuple.
        """

        if not isinstance(value, tuple):
            raise TypeError("Step arguments must be tuple")
        self.args = value

    def get_arguments(self):
        """Instruction action arguments getter.

        Returns:
            tuple: Instruction step arguments.
        """

        return self.args

    def set_kwarguments(self, value):
        """Instruction action keyword arguments setter.

        Args:
            value (dict): Instruction step keyword arguments.

        Raises:
            TypeError: If value is not dict.
        """

        if not isinstance(value, dict):
            raise TypeError("Step arguments must be dict")
        self.kwargs = value

    def get_kwarguments(self):
        """Instruction action keyword arguments getter.

        Returns:
            dict: Instruction step keyword arguments.
        """

        return self.kwargs

    def set_punctuation(self, value):
        """Instruction punctuation setter.

        Args:
            value (str): Instruction punctuation string.

        Raises:
            TypeError: If value is not a valid punctuation.
        """

        if value not in self.RunType.__dict__.values():
            raise TypeError("Unsupported step punctuation")
        self.punct = value

    def get_punctuation(self):
        """Instruction punctuation getter.

        Returns:
            Punctuation: Instruction punctuation flag.
        """

        return self.punct

    def run(self):
        """Run instruction.

        Raises:
            TypeError: If action is not callable or is not set.

        Returns:
            tuple: (Bool) True if succeeds, otherwise False; (String) output.
        """

        if not callable(self.action):
            raise TypeError("Action is not set")
        handler = self.action()
        handler.before_run()
        handler.set_payload(self.payload)
        if not isinstance(self.kwargs, dict):
            self.kwargs = {}
        if not isinstance(self.args, tuple):
            self.args = ()
        if len(self.kwargs) >= len(self.args):
            handler.run(**self.kwargs)
        else:
            handler.run(*self.args)
        handler.after_run()
        self.status, self.output = handler.get_status()
        return self.status, self.output

    def __repr__(self):
        return unicode(self.__dict__)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import chmod, getcwd, path as fpath

from buildok.action import Action


class ChangeMod(Action):
    r"""Change permissions on file or directory.

    Args:
        mode (str): Octal integer permissions.
        path (str): Path to file or directory.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `path` is provided.
        TypeError: If an invalid `mode` is provided.

    Accepted statements:
        ^change permissions to `(?P<mode>.+)`$
        ^change permissions to `(?P<mode>.+)` for `(?P<path>.+)`$
        ^change permissions `(?P<mode>.+)` for `(?P<path>.+)`$
        ^set permissions (?:to )?`(?P<mode>.+)` (?:for|to|on) `(?P<path>.+)`$

    Sample input:
        - Run `touch /tmp/buildok_test.txt`.
        - Set permissions to `400` for `/tmp/buildok_test.txt`.

    Expected:
        Changed permissions 400 => /tmp/buildok_test.txt
    """

    def run(self, mode="400", path=None, *args, **kwargs):
        if path is None:
            path = getcwd()
        try:
            chmod(path, int(mode, 8))
            self.success("Changed permissions %s => %s" % (mode, path))
        except OSError as e:
            self.fail(str(e))
        except TypeError as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, mode="400", path=None, *args, **kwargs):
        if path is None:
            path = "."
        flags = kwargs.get("flags", "")
        if fpath.isdir(path):
            flags = " -R"
        return "chmod%s %s %s" % (flags, mode, path)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from os import listdir
from platform import system

from buildok import __build__


DISCLAIMER = r"""
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class Sysenv(object):
    """System environment wrapper.

    Used in various files as a bash ENV-like to determine OS-related calls.

    app_version  (str): Application version (v0.0.0).
    os_name      (str): Operating System name.
    os_version   (str): Operating System version.
    shell_args (Shell): Shell instance with arguments.
    """

    app_version = ""
    os_name = ""
    os_version = ""
    shell_args = None

    @classmethod
    def print_disclaimer(cls):
        """Outputs application disclaimer.
        """

        print(DISCLAIMER)

    @classmethod
    def print_support(cls):
        """Outputs support message.
        """

        print("Please report bugs at https://github.com/lexndru/buildok")
        print("Use --help to see a list of all options")

    @classmethod
    def print_version(cls):
        """Outputs application version.
        """

        v_app, v_os = cls.app_version, cls.os_version
        print("Buildok {} [build {}] {}".format(v_app, __build__, v_os))

    @classmethod
    def print_headers(cls):
        """Output application version, disclaimer and support.
        """

        # Print version
        cls.print_version()
        print("")

        # Print disclaimer
        cls.print_disclaimer()
        print("")

        # Print support
        cls.print_support()
        print("")

    @classmethod
    def setup(cls, version, args):
        """Setup system environment.
        """

        cls.shell_args = args

        # Exit if version is missing
        if not version:
            raise SystemExit("Invalid launch")

        # Set app version
        cls.app_version = version

        # Buildok version header
        cls.os_version = ""

        # Basic support for windows systems
        if system().lower() == "windows":
            cls.os_version = u"WINDOWS"
            cls.os_name = "win"

        # Extended support for macos systems
        elif system().lower() == "darwin":
            cls.os_version = u"MACINTOSH"
            cls.os_name = "mac"

        # Full support for linux systems
        elif system().lower() == "linux":
            files = [f for f in listdir("/etc") if f.endswith("-release")]
            distro, pretty_name = "", ""
            for each_file in files:
                data = []
                with open("/etc/%s" % each_file, "r") as fd:
                    data = fd.readlines()
                for line in data:
                    line_ = line.lower()
                    if line_.startswith("id") and distro == "":
                        _, distro = line.split("=", 1)
                    if line_.startswith("pretty_name") and pretty_name == "":
                        _, pretty_name = line.split("=", 1)
            os_name = distro.strip()
            clean_name = pretty_name.strip()
            cls.os_version = u"%s (%s)" % (os_name.upper(), clean_name)
            cls.os_name = os_name

        # Extended or full support for bsd systems
        elif "bsd" in system().lower():
            cls.os_version = "BSD"
            cls.os_name = "bsd"



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.action import Action


class Touch(Action):
    r"""Create a new file.

    Args:
        filepath (str): Full path of the file.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `filepath` is provided.

    Accepted statements:
        ^create(?: new)? file `(?P<filepath>[\w\.]+)`$

    Sample input:
        - Create file `/tmp/buildok.txt`.

    Expected:
        Created file => /tmp/buildok.txt
    """

    def run(self, filepath=None, *args, **kwargs):
        try:
            with open(filepath, "w") as file_:
                file_.write("")
            self.success("Created file => %s" % filepath)
        except Exception as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, filepath=None, *args, **kwargs):
        if filepath is None:
            return "echo invalid touch call"
        return "touch %s" % filepath



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from buildok.util.log import Log


class UnicodeIcon(object):

    VALID = u"\033[92m\u2713\033[0m"
    INVALID = u"\033[91m\u237B\033[0m"


def crash(message):
    """Crash application with message.

    Raises:
        SystemExit.
    """

    return Log.fatal("Unexpected crash! %s" % message)


def scan_statements(statement):
    """Scans all statements.

    Returns:
        dict: Dictionary of all statements with a counter.
    """

    lines = {}
    if statement is None:
        return lines
    for action in statement.get_actions():
        for line in action.parse_statements():
            if lines.get(line) is None:
                lines.update({line: 0})
            lines[line] += 1
    return lines


def self_analyze(lines=None, statement=None):
    """Self analyze for errors.

    Returns:
        bool: True if all statements are ok, otherwise False.
    """

    if lines is None:
        lines = scan_statements(statement)
    return all([v == 1 for v in lines.itervalues()])


def fmt(fmt, txt):
    """Text formatter helper.
    """

    return fmt % txt


def analyze(statement, length=80):
    """Analyze all statements and print visual results.

    Returns:
        bool: True if all statements are ok, otherwise False.
    """

    lines = scan_statements(statement)
    results = self_analyze(lines)
    Log.debug("Listing all statements...")
    for action in statement.get_actions():
        try:
            text = action.parse_description()
        except Exception:
            text = "no description"
        delimiter = "-" * (length-4)
        c_txt = u"\033[96m %-{}s \033[0m".format(length-6)
        r_txt = u"\033[91m %-{}s \033[0m".format(length-6)
        g_txt = u"\033[92m %-{}s \033[0m".format(length-6)
        print(u"\033[90m|---|%-{}s|\033[0m".format(length-4) % delimiter)
        print(u"\033[90m|   |\033[0m" + fmt(c_txt, text) + "\033[90m|\033[0m")
        print(u"\033[90m|---|%-{}s|\033[0m".format(length-4) % delimiter)
        for line in action.parse_statements():
            if lines[line] > 1:
                status = UnicodeIcon.INVALID
                line_text = fmt(r_txt, line.strip())
            else:
                status = UnicodeIcon.VALID
                line_text = fmt(g_txt, line.strip())
            print_line = u"\033[90m|\033[0m " + status + u" \033[90m|\033[0m"
            print_line += line_text + u"\033[90m|\033[0m"
            print(print_line)
    print(u"\033[90m|---|%-{}s|\033[0m".format(length-4) % delimiter)
    if not results:
        Log.error("Duplicated statements found!")
        for line, times in lines.iteritems():
            if times > 1:
                err = (times, line.strip())
                Log.error(" - line duplicated %d times: %s" % err)
        Log.error("Please correct these problems before running again.")
    else:
        Log.debug("Everything looks OK!")
    return results



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from argparse import ArgumentParser


class Shell(object):
    """Shell argument parser.

    Attributes:
        parser (ArgumentParser): Argument parser.
        args             (dict): Dictionary of arguments.
    """

    parser = ArgumentParser(description="A tool to automate build steps")
    args = {
        ("-g", "--guide"): {
            "action": "store",
            "dest": "guide",
            "help": "path to guide"
        },
        ("-v", "--verbose"): {
            "action": "store_true",
            "dest": "verbose",
            "help": "verbose output"
        },
        ("-c", "--convert"): {
            "action": "store",
            "dest": "convert",
            "help": "create automated build scripts from build steps"
        },
        ("-a", "--analyze"): {
            "action": "store_true",
            "dest": "analyze",
            "help": "scan all known statements"
        },
        ("-l", "--lookup"): {
            "action": "store",
            "dest": "lookup",
            "help": "analyze and lookup statement with example"
        },
        ("-t", "--topic"): {
            "action": "store",
            "dest": "topic",
            "help": "set topic to build"
        },
        (None, "--topic-pattern"): {
            "action": "store",
            "dest": "topic_pattern",
            "help": "set topic pattern"
        },
        ("-p", "--preview"): {
            "action": "store_true",
            "dest": "preview",
            "help": "visual preview of scanned guide"
        },
        (None, "--install-policy"): {
            "action": "store",
            "dest": "install_policy",
            "help": "set or install a system utility policy"
        },
        (None, "--fake-run"): {
            "action": "store_true",
            "dest": "fake_run",
            "help": "read and parse guide without actually running on machine"
        },
        (None, "--strict"): {
            "action": "store_true",
            "dest": "strict",
            "help": "don't ignore unsupported steps from guide"
        },
        (None, "--unsafe-shell"): {
            "action": "store_true",
            "dest": "unsafe_shell",
            "help": "spawn a process to inject shell commands provided by RUN"
        },
        (None, "--package-manager"): {
            "action": "store",
            "dest": "package_manager",
            "help": "set absolute path to package manager"
        },
        (None, "--version"): {
            "action": "store_true",
            "dest": "version",
            "help": "display version number"
        },
        (None, "--config-file"): {
            "action": "store",
            "dest": "config_file",
            "help": "set path to configuration file"
        },
        (None, "--placeholder"): {
            "action": "append",
            "dest": "placeholder",
            "help": "set value to replace placeholder"
        },
    }

    @classmethod
    def parse(cls):
        """Create parser listener.

        Return:
            Namespace: Namespace of parser arguments from shell.
        """

        for keys, vals in cls.args.iteritems():
            short, long_ = keys
            if short is None:
                cls.parser.add_argument(long_, **vals)
            else:
                cls.parser.add_argument(short, long_, **vals)
        return cls.parser.parse_args()



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from buildok.util.log import Log


class Report(object):
    """Report wrapper for automated scripts.

    Arguments:
        runtime      (int): Script runtime in seconds.
        status       (str): Human-readable text status (e.g. OK, Failed, Exit).
        topic        (str): Topic as text string.
        total_steps  (int): Total steps found in choosen topic.
        current_step (int): Last step counted while looping instructions.
        error        (str): Human-readable text error (e.g. Permission denied).
    """

    runtime = 0
    status = "n/a"
    topic = "n/a"
    total_steps = 0
    current_step = 0
    error = "n/a"

    @classmethod
    def set_runtime(cls, runtime):
        """Set runtime value.
        """

        cls.runtime = runtime

    @classmethod
    def set_status(cls, status):
        """Set status value.
        """

        cls.status = status

    @classmethod
    def set_topic(cls, topic):
        """Set topic value.
        """

        cls.topic = topic

    @classmethod
    def set_total_steps(cls, total_steps):
        """Set total steps value.
        """

        cls.total_steps = total_steps

    @classmethod
    def set_error(cls, error):
        """Set error value.
        """

        cls.error = error

    @classmethod
    def inc_step(cls, inc):
        """Increment current step value.
        """

        cls.current_step += inc

    @classmethod
    def output(cls):
        """Dump report of each saved property.
        """

        ran_status = (cls.current_step, cls.total_steps)
        Log.info("""Topic name: "%s" """ % cls.topic)
        Log.info("Steps ran successful %d out of %d" % ran_status)
        Log.info("""Last step error: "%s" """ % cls.error)
        Log.info("Runtime %ss" % cls.runtime)
        Log.info("Build %s" % cls.status.upper())



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging


class Log(object):
    """Simple log wrapper
    """

    config = {
        "format": u"%(levelname)8s %(asctime)-15s \033[90m|\033[0m %(message)s"
    }

    @classmethod
    def configure(cls, verbose=False):
        cls.config["level"] = logging.DEBUG if verbose else logging.INFO
        logging.basicConfig(**cls.config)

    @classmethod
    def info(cls, message):
        logging.info(message)

    @classmethod
    def debug(cls, message):
        logging.debug(message)

    @classmethod
    def warn(cls, message):
        logging.warning(message)

    @classmethod
    def error(cls, message):
        logging.error(message)

    @classmethod
    def fatal(cls, message):
        logging.critical(message)
        raise Exception(message)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from os import listdir, stat, getcwd, path as fpath
from stat import S_ISREG, S_ISFIFO, S_ISSOCK, S_ISLNK, \
                 S_ISDIR, S_ISCHR, S_ISBLK

from buildok.action import Action


class ListDir(Action):
    r"""List files in directory.

    Args:
        path (str): Path to directory.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `path` is provided.

    Accepted statements:
        ^list (?:files|folders) (?:in|from|of) `(?P<path>.+)`$
        ^list everything$

    Sample input:
        - List files in `/tmp`.

    Expected:
        Listing directory => /tmp
    """

    def run(self, path=None, *args, **kwargs):
        if path is None:
            path = getcwd()
        length = 80
        try:
            content = listdir(path)
            fullpath = fpath.abspath(path)
            self.success("Listing directory => %s" % fullpath)
            delimiter = "-" * (length)
            print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
            s_name = int((length-7) * .8)
            s_size = int((length-7) - s_name)
            row = u"\033[90m| %s |\033[0m"
            row += u" %-{}s\033[90m|\033[0m %-{}s".format(s_name, s_size)
            row += u"\033[90m|\033[0m"
            print(row % (" ", "Name", "Size"))
            print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
            for f in content:
                f_name = f if len(f) <= s_name else u"%s..." % f[:s_name-3]
                f_type = self.get_file_type(fullpath, f)
                if f_type == "f":
                    f_size = self.get_file_size(fullpath, f)
                else:
                    f_size = "--"
                print(row % (f_type, f_name, f_size))
            print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
        except Exception as e:
            self.fail(str(e))

    def get_file_size(self, root, filename):
        filepath = fpath.join(root, filename)
        return ListDir.sizeof_fmt(fpath.getsize(filepath))

    def get_file_type(self, root, filename):
        filepath = fpath.join(root, filename)
        filemode = stat(filepath).st_mode
        if S_ISREG(filemode):
            return "f"
        elif S_ISSOCK(filemode):
            return "s"
        elif S_ISFIFO(filemode):
            return "p"
        elif S_ISLNK(filemode):
            return "l"
        elif S_ISDIR(filemode):
            return "d"
        elif S_ISCHR(filemode):
            return "c"
        elif S_ISBLK(filemode):
            return "b"
        return "?"

    @classmethod
    def convert_shell(cls, path=None, *args, **kwargs):
        if path is None:
            path = "."
        return "ls -hal %s" % path

    @staticmethod
    def sizeof_fmt(num):
        # https://web.archive.org/web/20111010015624/http://blogmag.net/blog/read/38/Print_human_readable_file_size
        for x in ["bytes", "KB", "MB", "GB", "TB"]:
            if num < 1024.0:
                return "%3.1f%s" % (num, x)
            num /= 1024.0
        return num



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from os import environ

from buildok.version import __version__

from buildok.statement import Statement
from buildok.script import Script
from buildok.action import Action
from buildok.placeholder import Placeholder

from buildok.util.console import Console
from buildok.util.shell import Shell
from buildok.util.analyze import self_analyze, analyze, crash
from buildok.util.locker import lock, unlock
from buildok.util.lookup import scan_lookup
from buildok.util.sysenv import Sysenv
from buildok.util.log import Log


def main(error=None):

    # Parse command line args
    args = Shell.parse()

    # System setup
    Sysenv.setup(__version__, args)

    # Check version and exit
    if args.version:
        return Sysenv.print_version()

    # Output application intro headers
    Sysenv.print_headers()

    # Set log verbose level
    Log.configure(verbose=args.verbose)

    # Set console verbose level
    Console.verbose = args.verbose

    # Prepare all statements
    Statement.prepare()

    # Lookup statement and example usage
    if args.lookup is not None:
        return scan_lookup(Statement, args.lookup)

    # Run analysis and exit
    if args.analyze:
        return analyze(Statement)

    # Self-analyze buildok and continue or crash
    self_analyze(None, Statement) or crash("Run an analyze and fix problems")

    # Setup placeholders
    Placeholder.config(args.placeholder)

    # Attach system environment to all actions
    Action.set_env(Sysenv)

    # Initialize script and run all steps
    guide_script = Script(args)
    guide_script.setup()

    # Create a lock file
    lock()
    try:

        # Parse guide and run all steps
        guide_script.parse().run()

    except Exception as e:
        if str(e) != "":
            error = e

    # Free lock file
    unlock()

    # Display report
    guide_script.print_report()

    # Throw any errors found...
    if environ.get("DEBUG") is not None and error is not None:
        raise Console.fatal(error)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.action import Action


class InvokeTopic(Action):
    r"""Invoke another topic from guide.

    Args:
        topic (str): Topic complete name to invoke.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        Exception: If an invalid `topic` is provided.

    Accepted statements:
        ^read (?:steps from )?topic `(?P<topic>.+)`$
        ^continue reading topic `(?P<topic>.+)`$
        ^follow steps from `(?P<topic>.+)`$

    Sample input:
        - Follow steps from `do something else`.

    Expected:
        Running new topic => do something else
    """

    def run(self, topic=None, *args, **kwargs):
        pass



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import path, remove
from shutil import rmtree

from buildok.action import Action


class Remove(Action):
    r"""Remove files from a given source.

    Args:
        src (str): Source of files.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `src` is provided.

    Accepted statements:
        ^remove from `(?P<src>.+)`$
        ^remove `(?P<src>.+)` files$
        ^remove (?:file|folder|directory) `(?P<src>.+)`$

    Sample input:
        - Go to `/tmp`.
        - Run `touch buildok_test_tmp.txt`.
        - Remove file `buildok_test_tmp.txt`.

    Expected:
        Removed => buildok_test_tmp.txt
    """

    def run(self, src=None, *args, **kwargs):
        try:
            if path.isfile(src):
                remove(src)
            elif path.isdir(src):
                rmtree(src)
            self.success("Removed => %s" % src)
        except OSError as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, src=None, *args, **kwargs):
        flags = kwargs.get("flags", "")
        if src is None:
            src = "."
        if path.isdir(src):
            flags = " -r"
        if flags == "":
            flags = " -f"
        else:
            flags += "f"
        return "rm%s %s 2> /dev/null" % (flags, src)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import environ


class Placeholder(object):
    """Parse guide and replace palceholders with defined values.

    It works with an environment variable PLACEHOLDERS which stores the
    absolute path to a placeholder-holder file. Data stored inside the file
    has to follow the format:

        key1=value1
        key2=value2_value2.2
        keyX=768g2eyd921h81z

    Shell arguments can overwrite any key defined in placeholder file.
    Sample usage:

        --placeholder key1=newValue1 --placeholder key2=newPlaceholder

    Raises:
        Exception: If specified placeholder file is invalid.
    """

    storage = None

    @classmethod
    def config(cls, args):
        """Configure placeholder class instance.

        Args:
            args (list): Provided by shell; updates internal placeholder memory

        Raises:
            Exception: If file provided by environment variable is not valid
        """

        if not isinstance(cls.storage, dict):
            cls.storage = {}
        ph_file = environ.get("PLACEHOLDERS")
        if ph_file is not None:
            try:
                with open(ph_file) as fd:
                    data = fd.read().split("\n")
                    cls.storage.update(cls.parse_args(data))
            except Exception as e:
                raise SystemExit(str(e))
        cls.storage.update(cls.parse_args(args))

    @classmethod
    def parse_args(cls, args):
        """Arguments parser.

        Arguments must have the following format:

            key1=value1
            key2=value2

        Args:
            args (list): List of arguments to be parsed.

        Returns:
            dict: Dictionary of paired arguments as key-value.
        """

        if not isinstance(args, list):
            args = []
        return dict([tuple(kv.split("=", 1)) for kv in args if len(kv) > 1])

    @classmethod
    def parse_string(cls, string):
        """Key-value string parser.

        Replaces a string containing a placeholder with it's stored value.

        Args:
            string (str): String to scan and update.

        Returns:
            str: Updated string.
        """

        for k, v in cls.storage.iteritems():
            string = string.replace("<%s>" % k, v)
        return string

    @classmethod
    def scan_string(cls, string):
        """Placeholder scanner.

        Scan a string for placeholders.

        Args:
            string (str): String to scan.

        Returns:
            bool: True if string has placeholders.
        """

        for k in cls.storage.iterkeys():
            if "<%s>" % k in string:
                return True
        return False

    @classmethod
    def parse_list(cls, list_string):
        """Key-value list parser.

        Replaces a string containing a placeholder with it's stored value.

        Args:
            list_string (list): List of strings to scan and update.

        Returns:
            list: Updated list of strings.
        """

        return [cls.parse_string(s) for s in list_string]

    @classmethod
    def scan_list(cls, list_string):
        """Placeholder scanner.

        Scan a list of strings for placeholders.

        Args:
            list_string (list): List of strings to scan.

        Returns:
            bool: True if list has placeholders.
        """

        return any([cls.scan_string(s) for s in list_string])



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

import timeit
import sys
import traceback


def traceback_error():
    """For debug only. Traceback to error

    Outputs traceback stack to error.
    """
    try:
        _, _, tb = sys.exc_info()
        for line in traceback.extract_tb(tb):
            print(" ".join([str(e) for e in line]))
    except Exception:
        pass


class Console(object):
    """Console wrapper.

    Attributes:
        start_time (int): Start time of console logging.
        stop_time  (int): Stop time of console logging.
        verbose   (bool): Set verbose level on or off.
    """
    start_time, stop_time = 0, 0
    verbose = False

    @classmethod
    def darkgray(cls, text, *args):
        """Print text in dark gray color.

        Return:
            str: Colored text.
        """
        return u"\033[90m{}\033[0m {}".format(text, " ".join(args))

    @classmethod
    def green(cls, text, *args):
        """Print text in green color.

        Return:
            str: Colored text.
        """
        return u"\033[92m{}\033[0m {}".format(text, " ".join(args))

    @classmethod
    def cyan(cls, text, *args):
        """Print text in cyan color.

        Return:
            str: Colored text.
        """
        return u"\033[96m{}\033[0m {}".format(text, " ".join(args))

    @classmethod
    def yellow(cls, text, *args):
        """Print text in yellow color.

        Return:
            str: Colored text.
        """
        return u"\033[93m{}\033[0m {}".format(text, " ".join(args))

    @classmethod
    def red(cls, text, *args):
        """Print text in red color.

        Return:
            str: Colored text.
        """
        return u"\033[91m{}\033[0m {}".format(text, " ".join(args))

    @classmethod
    def log(cls, message=None):
        """Print log message.

        Args:
            message (str): Optional message to display.
        """
        print("[Log] %s" % message)

    @classmethod
    def start(cls, message=None):
        """Start console timer.

        Args:
            message (str): Optional message to display on start-up.
        """
        cls.start_time = timeit.default_timer()
        if cls.verbose and message is not None:
            print("[Build] %s" % message)

    @classmethod
    def stop(cls, message=None):
        """Stop console timer.

        Args:
            message (str): Optional message to display on end.
        """
        cls.stop_time = timeit.default_timer()
        if cls.verbose and message is not None:
            print("\033[92m[Build] %s\033[0m" % message)
        seconds = cls.stop_time - cls.start_time
        print("\033[92m\033[1m[Build] OK (runtime %ss)\033[0m" % seconds)

    @classmethod
    def warn(cls, message):
        """Console simple warn message.

        Args:
            message (str): Message to display.
        """
        if cls.verbose:
            print(u"\033[93m[Build] %s\033[0m" % message)

    @classmethod
    def info(cls, message):
        """Console simple info message.

        Args:
            message (str): Message to display.
        """
        if cls.verbose:
            print(u"\033[92m[Build] %s\033[0m" % message)

    @classmethod
    def eval(cls, message):
        """Console response message.

        Args:
            message (str): Message to display.
        """
        if cls.verbose:
            print(u"\033[93m[Yield] %s\033[0m" % message)

    class fatal(SystemExit):
        """Console fatal error message.

        Args:
            err (str): Error to display.
        """
        def __init__(self, err):
            error = u"\033[91m[Fatal] %s\033[0m" % unicode(err)
            super(self.__class__, self).__init__(error)
            traceback_error()

    class debug(Exception):
        """Console debug error message.

        Args:
            err (str): Error to display.
        """
        def __init__(self, err):
            error = u"\033[93m[Error] %s\033[0m" % unicode(err)
            super(self.__class__, self).__init__(error)
            traceback_error()


def timeit_log(func):
    def wrapper(*args, **kwargs):
        Console.start()
        retval = func(*args, **kwargs)
        Console.stop()
        return retval
    return wrapper



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import kill
from signal import SIGTERM
from subprocess import check_output, CalledProcessError

from buildok.action import Action


class KillProcess(Action):
    r"""Send SIGTERM signal to a process.

    Args:
        pid (int): Process ID.
        pname (str): Process name.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `pid` or `pname` is provided.
        CalledProcessError: If `pname` is not found.

    Accepted statements:
        ^(?:kill|stop) process `(?P<pname>.+)`$
        ^(?:kill|stop) pid `(?P<pid>.+)`$

    Sample (input):
        - Stop process `someProcessName`.

    Expected:
        Terminated process PID => 9999
    """

    def run(self, pid=None, pname=None, *args, **kwargs):
        try:
            if pname is not None:
                pid = check_output(["pidof", "-s", pname])
            if pid is None:
                raise ValueError("Invalid PID")
            kill(int(pid), SIGTERM)
            self.susccess("Terminated process PID => %s" % pid)
        except OSError as e:
            self.fail(str(e))
        except ValueError as e:
            self.fail(str(e))
        except CalledProcessError as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, pid=None, pname=None, *args, **kwargs):
        if pid is not None:
            return "kill %s" % pid
        elif pname is not None:
            return "kill `pidof %s`" % pname
        return "echo cannot kill process because of invalid or missing pid"



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from shlex import split as cmd_split
from subprocess import check_output, CalledProcessError, STDOUT

from buildok.action import Action


class ShellExec(Action):
    r"""Run a command in shell.

    Args:
        cmd (str): Raw shell command.

    Retuns:
        str: Output as string.

    Raises:
        OSError: If an invalid `cmd` is provided.

    Accepted statements:
        ^run `(?P<cmd>.+)`$

    Sample input:
        - Run `echo hello friend how are you`.

    Expected:
        Output => hello friend how are you
    """

    def run(self, cmd=None, *args, **kwargs):
        safe_cmd = cmd_split(cmd)
        output_args = {"stderr": STDOUT}
        if self.env.shell_args is not None:
            if self.env.shell_args.unsafe_shell:
                output_args.update({"shell": True})
            else:
                cmd = safe_cmd
        try:
            output = check_output(cmd, **output_args)
            if output is None:
                output = "n/a"
            self.success(u"Output => %s" % output.decode('utf-8').strip())
        except CalledProcessError as e:
            self.fail(e.output)
        except Exception as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, cmd=None, *args, **kwargs):
        if cmd is None:
            return "echo Invalid script or missing cmd"
        return cmd



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class Guide(object):
    """A guide is a collection of topics.

    Class arguments:
        __guides (list): Private list of guides.

    Args:
        topics (tuple): Imutable list of topics.
    """

    __guides = []

    def __init__(self, topics=None):
        self.topics = topics
        self.__guides.append(self)

    def add_topic(self, topic):
        """Appends topic to guide instance.

        Args:
            topic (str): Topic managed by current guide.
        """
        if self.topics is None:
            self.topics = []
        self.topics.append(topic)

    def set_topics(self, value):
        """Guide setter for topics list.

        Args:
            value (list): Topics list.

        Raises:
            TypeError: If value is not a valid list.
        """
        if not isinstance(value, list):
            raise TypeError("Unsupported topics list")
        self.topics = value

    def get_topics(self):
        """Guide getter for topics list.

        Returns:
            list: List of topics.
        """
        return self.topics

    def get_topic_by_title(self, title):
        """Lookup topic by title and return instance.

        Returns:
            mixt: Instance of topic or None.
        """
        for t in self.topics:
            if t.get_title().lower() == title.lower():
                return t
        return None

    @classmethod
    def get_all_guides(cls):
        """Return stored guide instances.

        Returns:
            list: List of stored guides.
        """
        return cls.__topics

    def __repr__(self):
        return unicode(self.__dict__)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from subprocess import Popen
from time import sleep
from shlex import split as cmd_split

from buildok.action import Action

from buildok.util.log import Log


class InstallPackage(Action):
    r"""Install new software package(s).

    Args:
        pkgs (str): Packages to install.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        Exception: If any of `pkgs` are not found in repositories.

    Accepted statements:
        ^install `(?P<pkgs>.+)`$

    Sample (input):
        - Install `vim curl`.

    Expected:
        Installed 2 new packages
    """

    os_packs = {
        ("alpine",):            "apk add {packages}",
        ("debian", "ubuntu"):   "apt-get install -y {packages}",
        # ("centos", "fedora"):   "yum install {packages}",
        # ("fedora",):            "dnf install -y {packages}",
        # ("slackware",):         "slackpkg install {packages}",
        # ("arch",):              "pacman -S {packages}",
        # ("gentoo",):            "emerge {packages}",
    }

    def run(self, pkgs=None, *args, **kwargs):
        packages = pkgs.split()
        if len(packages) == 0:
            return self.fail("No packages to install...")
        try:
            cmd = InstallPackage.check_install()
            if cmd is None:
                return self.fail("Unsupported OS: %s" % self.env.os_name)
            installed_pkgs = self.install_packages(cmd, packages)
            if installed_pkgs > 0:
                self.success("Installed %d new packages" % installed_pkgs)
            else:
                self.fail("Failed to install packages...")
        except Exception as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, pkgs=None, *args, **kwargs):
        if pkgs is None:
            return "echo Nothing to install"
        cmd = InstallPackage.check_install()
        if cmd is not None:
            return cmd.format(packages=pkgs)
        return "echo Unable to install packages: %s" % pkgs

    @classmethod
    def check_install(cls):
        for oses, cmd in cls.os_packs.iteritems():
            if cls.env.os_name in oses:
                return cmd
        return None

    def install_packages(self, cmd, packages):
        installed = 0
        for pkg in packages:
            install_cmd = cmd_split(cmd.format(packages=pkg))
            install_output = Popen(install_cmd)
            while install_output.poll() is None:
                sleep(0.5)
            output = install_output.returncode
            log_status = (self.env.os_name, install_cmd, output == 0)
            Log.debug("Running (%s) %s ... %r" % log_status)
            if 0 == output:
                installed += 1
        return installed



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from buildok.util.log import Log


def scan_lookup(statements, lookup, length=80):
    """Lookup a statement with example usage.

    Args:
        statements (Statement): Statement instance will all known statements.
        lookup           (str): String to lookup after in all statements.
    """

    Log.debug("Lookup after %s ..." % lookup)
    known_statements = statements.get_actions()
    actions = []

    # Check available statements
    if len(known_statements) == 0:
        return Log.error("Statements are missing")

    # Search in all statemnts
    for stmt in known_statements:
        if lookup.lower() in stmt.parse_description().lower():
            if stmt not in actions:
                actions.append(stmt)
        else:
            for exp in stmt.parse_statements():
                if lookup.lower() in exp.lower():
                    if stmt not in actions:
                        actions.append(stmt)

    # Handle missing action
    if len(actions) == 0:
        return Log.error("Nothing found for '%s'" % lookup)

    # Print lookup
    for action in actions:
        delimiter = "-" * (length)
        description = action.parse_description()
        desc = u"\033[96m %-{}s \033[0m".format(length-2) % description
        print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
        print(u"\033[90m|\033[0m" + desc + "\033[90m|\033[0m")
        print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
        for text in action.parse_statements():
            stmt = u"\033[95m %-{}s \033[0m".format(length-6) % text.strip()
            print(u"\033[90m|   |\033[0m" + stmt + "\033[90m|\033[0m")
        print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
        for text in action.parse("sample"):
            sample = u"\033[93m %-{}s \033[0m".format(length-6) % text.strip()
            print(u"\033[90m|   |\033[0m" + sample + "\033[90m|\033[0m")

    # Done
    print(u"\033[90m|%-{}s|\033[0m".format(length-4) % delimiter)
    Log.debug("Done...")



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import webbrowser as wb

from buildok.action import Action


class ViewWeb(Action):
    r"""Open a link in default browser.

    Args:
        url (str): URL to open.

    Retuns:
        str: Output as string.

    Raises:
        TypeError: If an invalid `url` is provided.

    Accepted statements:
        ^open in browser `(?P<url>.+)`$
        ^open (?:link|url) `(?P<url>.+)`$

    Sample (input):
        - Open link `https://github.com/lexndru/buildok`.

    Expected:
        Opened URL in browser => https://github.com/lexndru/buildok
    """

    def run(self, url=None, *args, **kwargs):
        error = self.open_url(url)
        if error is not None:
            self.fail(error)
        else:
            self.success("Opened URL in browser => %s" % url)

    def open_url(self, url):
        try:
            wb.get().open(url, new=2)
        except wb.Error as e:
            return str(e)
        except TypeError as e:
            return str(e)
        except Exception as e:
            return "Cannot open \"%s\": %s" % (url, str(e))
        return None

    @classmethod
    def convert_shell(cls, url=None, *args, **kwargs):
        if url is None:
            return "echo Cannot open browser: URL is missing?"
        return "echo Cannot open URL in browser: %s" % url



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.parser import Parser


class Action(object):
    """Action statement base handler.

    Arguments:
        header (str): Lookup docstring header.
        env (Sysenv): System environment instance.

    Args:
        output  (str): Output status message.
        failed (bool): Handler failture status.
        payload (str): Handler payload input message.

    Raises:
        any: Any exception raised by action handler
    """

    env = None
    doc_header = r"accepted statements"

    def __init__(self, payload=None):
        self.payload = payload
        self.output = ""
        self.failed = None

    @classmethod
    def set_env(cls, env):
        """System environment setter.

        Args:
            env (Sysenv): System environment instance.
        """

        if not callable(env):
            raise TypeError("Expected sysenv instance, got something else")
        cls.env = env

    def get_status(self):
        """Action status getter.

        Returns:
            tuple: Boolean if succeded or failed; and output message.
        """

        return (not self.failed, self.output)

    def get_output(self):
        """Output getter.

        Returns:
            str: Output message.
        """

        return self.output

    def set_output(self, message):
        """Output setter.

        Args:
            message (str): Output message.
        """

        self.output = message

    def get_payload(self):
        """Payload getter.

        Returns:
            str: Payload input message.
        """

        return self.payload

    def set_payload(self, payload):
        """Payload setter.

        Args:
            payload (str): Payload input message.
        """

        self.payload = payload

    def success(self, message=""):
        """Flag action as being successful.

        Args:
            message (str): Optional message to describe output.
        """

        self.failed = False
        self.output = message

    def fail(self, message=""):
        """Flag action as being failed.

        Args:
            message (str): Optional message to describe output.
        """

        self.failed = True
        self.output = message

    def has_failed(self):
        """Status getter.

        Returns:
            bool: True if failed, else False.
        """

        return self.failed

    def run(self, *args, **kwargs):
        """Action handler runnable.

        Runs the context of an action handler.

        Raises:
            NotImplementedError: If not implemented.
        """

        raise NotImplementedError("Class must implement this method")

    def before_run(self, *args, **kwargs):
        """Action handler prehook.

        Runs before an action handler.
        """

        pass

    def after_run(self, *args, **kwargs):
        """Action handler posthook.

        Runs after an action handler.
        """

        pass

    def test(self, *args, **kwargs):
        """Action handler testable.

        Tests the context of an action handler.

        Raises:
            NotImplementedError: If not implemented.
        """

        raise NotImplementedError("Class must implement this method")

    @classmethod
    def parse_description(cls, fallback_msg="No description"):
        """Action handler description.

        Returns:
            str: First line from docstring or fallback message.
        """

        try:
            return cls.__doc__.splitlines()[0]
        except IndexError:
            return fallback_msg

    @classmethod
    def parse_statements(cls):
        """Statements parser.

        Returns:
            list: List of statements extracted from action handler.
        """

        return cls.parse(cls.doc_header)

    @classmethod
    def parse(cls, string):
        """Docstring parser.

        Args:
            string (str): Headline string to lookup after.

        Returns:
            list: List extracted from action handler.
        """

        return Parser.lookahead(unicode(cls.__doc__), string)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import symlink, getcwd

from buildok.action import Action


class MakeSymlink(Action):
    r"""Make a symlink for a target source.

    Args:
        src (str): Source of files.
        dst (str): Target destination of symlink.

    Retuns:
        str: Human readable descriptor message or error.

    Raises:
        OSError: If an invalid `src` or `dst` is provided or if `dst` exists.

    Accepted statements:
        ^create symlink from `(?P<src>.+)` to `(?P<dst>.+)`$
        ^make symlink `(?P<dst>.+)` from `(?P<src>.+)`$
        ^make symlink `(?P<dst>.+)`$

    Sample input:
        - Run `touch buildok_test_symlink`.
        - Make symlink `buildok_test_symlink_ok` from `buildok_test_symlink`.

    Expected:
        Created symlink buildok_test_symlink => buildok_test_symlink_ok
    """

    def run(self, src=None, dst=None, *args, **kwargs):
        if src is None:
            src = getcwd()
        try:
            symlink(src, dst)
            self.success("Created symlink %s => %s" % (src, dst))
        except OSError as e:
            self.fail(str(e))

    @classmethod
    def convert_shell(cls, src=None, dst=None, *args, **kwargs):
        if src is None and dst is None:
            return "echo invalid ln command"
        elif src is None:
            src = "."
        elif dst is None:
            dst = "."
        return "ln -s %s %s" % (src, dst)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.statements.web import ViewWeb


class GoogleSearch(ViewWeb):
    r"""Open a Google search results in default browser.

    Args:
        search (str): Search string to lookup.

    Retuns:
        str: Output as string.

    Raises:
        TypeError: If an invalid `search` is provided.

    Accepted statements:
        ^google (?:for )?`(?P<search>.+)`$

    Sample (input):
        - Google `buildok`.

    Expected:
        Google search results => https://google.com/?q=buildok
    """

    def run(self, search=None, *args, **kwargs):
        url = r"https://www.google.com?q={}".format(search)
        error = self.open_url(url)
        if error is not None:
            self.fail(error)
        else:
            self.success("Google search results => %s" % url)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class Parser(object):
    """General purpose parser.

    Used to extract various text lines of statements from a provided context
    source. E.g. extract accepted statements from action handlers.
    """

    @classmethod
    def lookahead(cls, ctx, header, start_line=-1, first_only=True):
        """Context parser.

        Args:
            ctx (str or list): Context to scan.
            header      (str): Needle to lookup in context.
            start_line  (int): Context offset.

        Returns:
            list: List of statements extracted from context.
        """

        if isinstance(ctx, (str, unicode)):
            ctx = ctx.splitlines()
        for idx, line in enumerate(ctx):
            text = line.strip()
            if start_line > -1 and not text:
                return ctx[start_line+1:idx]
            if text.lower().startswith(header):
                if first_only and start_line > -1:
                    continue
                start_line = idx
        return []



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from os import path

from buildok.util.log import Log
from buildok.util.console import Console


class Reader(object):
    """Reader wrapper to gather build steps.

    Attributes:
        READER (str): Class extending Reader.
        PATH   (str): Path to build steps file.

    Args:
        validate (bool): Enable or disable build steps validation.
        filename  (str): Build filename.
        content  (list): Content from file context.
        cursor    (int): Current line index being scanned.
        line      (str): Current line content.
    """

    READER, PATH = None, r"."

    def __init__(self, validate=False):
        self.validate = validate
        self.filename = ""
        self.content, self.cursor, self.line = [], 0, ""
        self.setup() or self.crash()

    @classmethod
    def set_project_path(cls, project_path):
        """Change project path.

        Default is current working directory.

        Args:
            project_path (str): New project path.

        Raises:
            ValueError: If project path is not a string.
        """

        if not isinstance(project_path, (str, unicode)):
            Log.fatal("Project path must be string")
        cls.PATH = project_path.rstrip("/")

    def parse(self):
        """Parse build steps file.

        Raises:
            NotImplementedError: Readers must implement this method.
        """

        raise NotImplementedError("Must implement read method")

    def read(self):
        """Read build file content.

        Raises:
            OSError: If `filename` cannot be found or accesed.
        """

        with open(self.filename, "rb") as file_:
            self.content = [r.rstrip("\n") for r in file_.readlines()]

    def exists(self):
        """Check if build file exists.

        Returns:
            bool: Returns True if `filename` exists, otherwise False.
        """

        return path.isfile(self.filename)

    def crash(self):
        """Crash reader with an error.

        Raises:
            NotImplementedError: Reader must overwrite READER attribute.
        """

        raise NotImplementedError("Directly call on Reader not allowed")

    def setup(self):
        """Setup if reader is properly called.

        Raises:
            ValueError: Project does not have build file.

        Returns:
            bool: Returns True if project setup is done.
        """

        if self.READER is None or self.PATH is None:
            return False
        filepath = self.PATH
        if path.isdir(filepath):
            filepath = path.join(self.PATH, self.READER)
        self.filename = path.abspath(filepath)
        if not self.exists():
            Log.fatal("Project doesn't have proper build file: %s" % filepath)
        return True

    def get_build_source(self):
        """Get a tuple of build file and a validation flag.

        Returns:
            build_file (str): Build file.
            validate  (bool): True if has to be validated.
        """

        return (self.filename, self.validate)

    def has_next(self):
        """Check if reader has more lines to read.

        Returns:
            bool: True if it has unread lines, otherwise False.
        """

        more_lines = len(self.content) > self.cursor
        if more_lines:
            self.line = self.content[self.cursor]
        return more_lines

    def get_line(self, strip=False):
        """Get current line read.

        Args:
            strip (bool): Strips text of whitespace if True.

        Returns:
            str: Current line from build file.
        """

        return self.line.strip() if strip else self.line

    def get_line_number(self):
        """Get current line number.

        Returns:
            int: Current line number from build file.
        """

        return self.cursor

    def next_line(self):
        """Go to next line in build file.

        It actually calls skip_line() with 1 iteration.
        """

        self.skip_line()

    def skip_line(self, lines=1):
        """Jump X lines ahead in build file.

        Move line cursor ahead. By default it jumps one line.
        Can go backwards if negative number is provided.
        """

        self.cursor += lines

    def next_content(self):
        """Get content ahead of cursor.

        Returns:
            list (list): Returns content ahead of cursor or empty list.
        """

        index = self.cursor + 1
        if index >= len(self.content):
            return []
        return self.content[index:]

    def preview(self, limit=80):
        """Preview build file.

        Outputs list of all lines from build file with line number in front.
        """

        self.read()
        max_line = len(str(len(self.content))) + 1
        limit -= max_line
        print()
        for i, line in enumerate(self.content):
            _line = line
            preview_line = ""
            prefix = ("%" + str(max_line) + "d |") % (i + 1)
            if len(line) == 0:
                preview_line = prefix + " [newline]"
            while len(line) > 0:
                preview_line += "%s %s\n" % (prefix, line[:limit])
                line = line[limit:]
            print(self.preview_line(_line, preview_line.rstrip("\n")))
        print()

    def preview_line(self, line, preview_line):
        """Preview line from build file.

        Useful to customize the display row of a line.
        """

        return Console.darkgray(preview_line)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from shlex import split as cmd_split
from subprocess import Popen, CalledProcessError
from time import sleep

from buildok.action import Action

from buildok.util.log import Log


class EnableService(Action):
    r"""Enable service at boot time.

    Args:
        srv (str): Service name to enable.

    Retuns:
        str: Output as string.

    Raises:
        OSError: If an invalid `srv` is provided.

    Accepted statements:
        ^enable service `(?P<srv>.+)`$

    Sample (input):
        - Enable service `urandom`.

    Expected:
        Service 'urandom' => enabled
    """

    os_distro_trigger = {
        ("arch",
         "centos",
         "debian",
         "fedora",
         "gentoo",
         "ubuntu"): ("systemctl enable {service}.service",
                     "systemctl is-enabled {service}.service")
    }

    def run(self, srv=None, *args, **kwargs):
        cmd = EnableService.check_systemd()
        if cmd is None:
            return self.fail("Unsupported OS: %s" % self.env.os_name)
        try:
            toggle_cmd, check_cmd = cmd
            service_cmd = toggle_cmd.format(service=srv)
            log_status = (self.env.os_name, service_cmd)
            Log.debug("Service OS (%s) boot: %s ..." % log_status)
            service_output = Popen(cmd_split(service_cmd))
            while service_output.poll() is None:
                sleep(0.5)
            if 0 != service_output.returncode:
                return self.fail(u"Service '%s' => failed to enable" % srv)
            self.success(u"Service '%s' => enabled" % srv)
        except CalledProcessError as e:
            self.fail(e.output)
        except Exception as e:
            self.fail(str(e))

    @classmethod
    def check_systemd(cls):
        for oses, cmds in cls.os_distro_trigger.iteritems():
            if cls.env.os_name in oses:
                return cmds
        return None

    @classmethod
    def convert_shell(cls, srv=None, *args, **kwargs):
        if srv is None:
            return "echo Invalid script or missing service"
        cmd, _ = cls.check_systemd()
        if cmd is not None:
            return cmd.format(service=srv)
        return "echo Unable to get status of service %s" % srv



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.statement import Statement


class Matcher(object):
    """A small wrapper to pair statements and steps.

    Requires statements to be processed before running matching algorithm.
    """

    @classmethod
    def is_valid(cls, text_step):
        """Check if step is supported by statements.

        Args:
            text_step (str): Text string of step to check.

        Returns:
            bool: True if statements have support for step.
        """

        step = text_step.strip()
        if len(step) == 0:
            return False
        for exp, _ in Statement.get_statements():
            if exp.match(step) is not None:
                return True
        return False

    @classmethod
    def pair_all(cls, instructions):
        """Loop through all instructions and pair each step.

        Args:
            instructions (list): List of instructions.

        Returns:
            bool: True if all steps have been paired successfully.
        """

        return all([cls.pair_one(i) for i in instructions])

    @classmethod
    def pair_one(cls, instruction):
        """Pair instruction with statement.

        Args:
            instruction (Instruction): An instruction instance to pair.

        Returns:
            bool: True if instruction has a pair whitin statements.
        """

        for exp, fun in Statement.get_statements():
            args = exp.match(instruction.get_step())
            if args is not None:
                instruction.set_description(fun.parse_description())
                instruction.set_statement(fun)
                instruction.set_arguments(args.groups())
                instruction.set_kwarguments(args.groupdict())
                return True
        return False



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from re import compile, IGNORECASE, UNICODE

from buildok.statements.chdir import ChangeDir
from buildok.statements.mkdir import MakeDir
from buildok.statements.symlink import MakeSymlink
from buildok.statements.web import ViewWeb
from buildok.statements.google import GoogleSearch
from buildok.statements.duckduckgo import DuckDuckGoSearch
from buildok.statements.wikipedia import WikipediaSearch
from buildok.statements.github_search import GitHubSearch
from buildok.statements.shell import ShellExec
from buildok.statements.chmod import ChangeMod
from buildok.statements.chown import ChangeOwner
from buildok.statements.kill import KillProcess
from buildok.statements.copy import Copy
from buildok.statements.move import Move
from buildok.statements.remove import Remove
from buildok.statements.touch import Touch
from buildok.statements.edit_file import EditFile
from buildok.statements.install_pip import PipInstallPackage
from buildok.statements.install_npm import NpmInstallPackage
from buildok.statements.install import InstallPackage
from buildok.statements.uninstall import UninstallPackage
from buildok.statements.reinstall import ReinstallPackage
from buildok.statements.service_enable import EnableService
from buildok.statements.service_disable import DisableService
from buildok.statements.service_status import StatusService
from buildok.statements.service_start import StartService
from buildok.statements.service_stop import StopService
from buildok.statements.service_restart import RestartService
from buildok.statements.service_reload import ReloadService
from buildok.statements.listdir import ListDir
from buildok.statements.invoke import InvokeTopic
from buildok.statements.noop import Noop

from buildok.util.log import Log


class Statement(object):
    """Statement parser and launcher.

    Attributes:
        actions (frozen set): Set of all known statements and actions.
        statements    (list): List of all statements.
        ready         (bool): Setup flag to determine status.
    """

    __actions = {
        ChangeDir,          # Change current working directory.
        MakeDir,            # Make a directory or make recursive directories.
        MakeSymlink,        # Make a symlink for a target source
        ViewWeb,            # Open a link in default browser.
        GoogleSearch,       # Perform a Google search and open default browser.
        DuckDuckGoSearch,   # Perform a DuckDuckGo search and open browser.
        WikipediaSearch,    # Perform a Wikipedia search and open browser.
        GitHubSearch,       # Open a GitHub search in default browser.
        ShellExec,          # Run a command in shell.
        ChangeMod,          # Change permissions on file or directory.
        ChangeOwner,        # Change owner and group on file or directory.
        Copy,               # Copy files from source to destination.
        Move,               # Move files from source to destination.
        Remove,             # Remove files from a given source.
        KillProcess,        # Send SIGTERM signal to a process.
        Touch,              # Create a new file.
        EditFile,           # Edit content of an existing file.
        PipInstallPackage,  # Install Python packages.
        NpmInstallPackage,  # Install Node.js packages.
        InstallPackage,     # Install new package software.
        UninstallPackage,   # Uninstall package software.
        ReinstallPackage,   # Reinstall package software.
        InvokeTopic,        # Invoke new topic from guide.
        Noop,               # No operation; nothing to do.
        EnableService,      # Enable service at boot time.
        DisableService,     # Disable service at boot time.
        StatusService,      # Get status of service.
        StartService,       # Start new service.
        StopService,        # Stop running service.
        RestartService,     # Restart running service.
        ReloadService,      # Reload service configuration.
        ListDir,            # List files and directories from directory.
    }

    statements = {}
    ready = False

    @classmethod
    def prepare(cls):
        """Statement initialization.

        Loops through all supported actions, validates and maps statements with
        actions.

        Returns:
            bool: True if statements are mapped successful.
        """

        Log.debug("Preparing to scan %d actions" % len(cls.__actions))
        for action in cls.__actions:
            if not callable(action):
                raise SystemExit("Expected action to be callable")
            Log.debug("Scanning action: %s" % action.parse_description())
            for line in action.parse_statements():
                exp = compile(line.strip(), IGNORECASE | UNICODE)
                cls.statements.update({exp: action})
                Log.debug("Updating known patterns: %s" % exp.pattern)
        if len(cls.statements) < len(cls.__actions):
            raise SystemExit("Unable to map statements to action")
        cls.ready = True
        Log.debug("All statements are scanned")

    @classmethod
    def find_statement(cls, stmt):
        """Lookup a statement.

        Returns:
            mixt: Statement if found, otherwise None.
        """

        return cls.statements.get(stmt)

    @classmethod
    def has_statement(cls, stmt):
        """Check whetever a statement exists.

        Returns:
            bool: True if statement is found.
        """

        return cls.find_statement(stmt) is not None

    @classmethod
    def get_statements(cls):
        """Statetements getter.

        Returns:
            iterator: A key-value itertator of all statements.
        """

        return cls.statements.iteritems()

    @classmethod
    def get_actions(cls):
        """Actions getter.

        Returns:
            list: All supported actions.
        """

        return cls.__actions



#!/usr/bin/env python

from setuptools import setup


long_description = """
A tool to automate build steps from README files.
"""

setup(name="buildok",
    packages=[
        "buildok",
        "buildok.readers",
        "buildok.statements",
        "buildok.structures",
        "buildok.util",
        "buildok.converters",
    ],
    entry_points = {
        "console_scripts": [
            "build = buildok.bootstrap:main"
        ]
    },
    version="0.4.2",
    description="A tool to automate build steps from README files.",
    long_description=long_description,
    author="Alexandru Catrina",
    author_email="alex@codeissues.net",
    license="MIT",
    url="https://github.com/lexndru/buildok",
    download_url="https://github.com/lexndru/buildok/archive/v0.4.2.tar.gz",
    keywords=["build-tool", "build-automation", "readme", "buildok"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
    ],
)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from buildok.action import Action
from buildok.parser import Parser
from buildok.statement import Statement
from buildok.util.sysenv import Sysenv
from buildok.structures.instruction import Instruction


def main():

    # Load statemets
    Statement.prepare()

    # Attach sysenv to actions
    Action.set_env(Sysenv)

    # Holder
    not_testing = []

    # Loop all actions
    for action in Statement.get_actions():

        # Get name and description
        action_name = action.__name__
        action_desc = action.parse_description()

        # Get input and output
        data_in = Parser.lookahead(unicode(action.__doc__), "sample input")
        data_out = Parser.lookahead(unicode(action.__doc__), "expected")

        # Avoid actions with no tests
        if len(data_in) == 0:
            not_testing.append((action_name, action_desc))
            continue

        # Validate action
        print("Testing... %s (%s)" % (action_name, action_desc))

        for line in data_in:
            print(" " * 10, line.strip())

            scan = Instruction.PATTERN.match(line.strip())
            if scan is None:
                continue
            step = Instruction(0, scan.group("step"), scan.group("punct"))

            for exp, handler in Statement.get_statements():
                args = exp.match(step.get_step())
                if args is not None:
                    fun = handler()
                    fun.run(**args.groupdict())
                    print(fun.get_output())
                    break

        for line in data_out:
            print("Results...", line.strip())
        print("\n")

    # Print actions with no tests
    print("-" * 10, "\nNot tested...")
    for each in not_testing:
        print(" - %s (%s)" % each)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os

from buildok.converters.bash import BashConverter


class Converter(object):
    """Translate build steps to script files.

    Used as a converter wrapper for multiple scriping files.

    Args:
        workdir     (str): Working directory.
        conv_target (str): Choosen target to convert.
        tpl_target  (str): Unpacked holder for script parameters.
    """

    target_name, target_class = None, None
    workdir = os.getcwd()
    prefix = "convert_"

    @classmethod
    def prepare(cls, target):
        if target == "shell" or target == "bash":
            cls.target_class = BashConverter
            cls.target_name = "shell"
        elif target == "vagrant":
            raise Exception("Unsupported yet: %s" % target)
        elif target == "docker":
            raise Exception("Unsupported yet: %s" % target)
        elif target == "jenkins":
            raise Exception("Unsupported yet: %s" % target)
        elif target == "ansible":
            raise Exception("Unsupported yet: %s" % target)
        else:
            raise Exception("Unsupported target: %s" % target)

    @classmethod
    def check(cls):
        return cls.target_name is not None and cls.target_class is not None

    @classmethod
    def save(cls, steps):
        lines = []
        method = "%s%s" % (cls.prefix, cls.target_name)
        for step in steps:
            if not hasattr(step.action, method):
                continue
            if not isinstance(step.args, tuple):
                step.args = ()
            func = getattr(step.action, method)
            comment = "\n# %s" % step.step
            cmdline = func(*step.args, payload=step.payload)
            lines.append("%s\n%s" % (comment, cmdline))
        filepath = os.path.join(cls.workdir, cls.target_class.filename)
        with open(filepath, "w") as file_:
            data = cls.target_class.template.format(body="\n".join(lines))
            file_.write(data)



# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from buildok.reader import Reader
from buildok.matcher import Matcher
from buildok.placeholder import Placeholder

from buildok.structures.instruction import Instruction
from buildok.structures.guide import Guide
from buildok.structures.topic import Topic

from buildok.util.console import Console
from buildok.util.readerr import ReadError
from buildok.util.log import Log


class ReadmeReader(Reader):
    """File class helper to detect and read a README file.

    Class arguments:
        last_guide      (Guide): Last Guide instance.
        last_topic      (Topic): Last Topic instance.
        last_step (Instruction): Last Instruction instance.
        delimiter         (str): Payload delimiter.
        recent_topic      (str): Holder for recent topic.
        no_topic_skip    (bool): Skip instruction if topic is missing.
    """

    READER = r"README.md"

    class NoPayloadError(ReadError):

        error = r"Instruction payload missing for step: %s"

    class UnclosedPayloadError(ReadError):

        error = r"Unclosed instruction payload for step: %s"

    class BadFormatError(ReadError):

        error = r"Missing instruction payload for step: %s"

    last_guide = None
    last_topic = None
    last_step = None

    recent_topic = "n/a"
    no_topic_skip = True

    delimiter = r"```"
    __topicre = Topic.prepare(r"## how to (?P<topic>[\s\w\-]+)")

    def parse(self, newlines=0):
        """Parse build steps file.

        Loop through all steps and check for topics and instructions.
        """

        Log.debug("Guide has %d lines" % len(self.content))
        self.last_guide = Guide()
        while self.has_next():
            line = self.get_line(strip=True)
            if len(line) == 0:
                newlines += 1
            else:
                newlines = 0
            if newlines >= 2:
                self.last_topic = None
            self.check_topic(line)
            self.check_instruction(line).check_payload(self.next_content())
            self.next_line()

    def get_guide(self):
        """Get latest guide.

        Returns:
            mixt: Guide instance if is set, otherwise None.
        """

        return self.last_guide

    def get_guide_by_topic(self):
        """Get new guide filtered by topic.

        Returns:
            mixt: Guide instance if is set, otherwise None.
        """

        for topic in self.last_guide.get_topics():
            if Topic.TOPIC == topic.get_title():
                guide = Guide()
                guide.add_topic(topic)
                return guide
        return None

    def check_topic(self, line):
        """Check line from build steps if it's a topic.

        Args:
            line (str): Build step line as string.

        Returns:
            self: Self instance.
        """

        scan = Topic.PATTERN.match(line)
        if scan is not None:
            title = scan.group("topic")
            self.last_topic = Topic(self.get_line_number(), title)
            self.last_guide.add_topic(self.last_topic)
        return self

    def check_instruction(self, line):
        """Check line from build steps if it's an instruction.

        Args:
            line (str): Build step line as string.

        Raises:
            Exception: If last topic is missing and step is found.

        Returns:
            self: Self instance.
        """

        if self.no_topic_skip and self.last_topic is None:
            self.last_step = None
            return self
        scan = Instruction.PATTERN.match(line)
        if scan is not None:
            step = scan.group("step")
            if Placeholder.scan_string(step):
                step = Placeholder.parse_string(step)
            punct = scan.group("punct")
            self.last_step = Instruction(self.get_line_number(), step, punct)
            if self.last_topic is None:
                Log.fatal("Unaccepted instruction: missing topic in guide")
            self.last_topic.add_step(self.last_step)
        return self

    def check_payload(self, content):
        """Check line from build steps if it's an instruction payload.

        Args:
            content (list): Build step lines ahead of current cursor.

        Returns:
            self: Self instance.
        """

        if self.last_step is None:
            return self
        if self.last_step.punct != Instruction.RunType.ARGS:
            return self
        newlines, borders = 0, []
        for i, next_line in enumerate(content):
            if len(borders) == 0 and newlines > 1:
                raise self.NoPayloadError(self.last_step)
            if next_line.strip() == "":
                newlines += 1
                continue
            if next_line.strip() == self.delimiter:
                borders.append(i)
            if len(borders) == 2:
                break
        if len(borders) == 1:
            raise self.UnclosedPayloadError(self.last_step)
        try:
            a, z = borders
            payload = content[1+a:z]
            if Placeholder.scan_list(payload):
                payload = Placeholder.parse_list(payload)
            self.last_step.set_payload(payload)
            self.skip_line(lines=z+1)
            self.last_step = None
        except Exception as e:
            Log.error(str(e))
            raise self.BadFormatError(self.last_step)
        return self

    def preview_line(self, line, preview_line):
        """Preview line from build file.

        Useful to customize the display row of a line.
        """

        line_topic = Topic.PATTERN.match(line)
        if line_topic is not None:
            self.recent_topic = line_topic.group("topic")
            line_desc = "topic (%s)" % self.recent_topic
            return Console.green(preview_line, "<--- %s" % line_desc)
        line_step = Instruction.PATTERN.match(line)
        if line_step is not None:
            step = line_step.group("step")
            if Placeholder.scan_string(step):
                step = Placeholder.parse_string(step)
            topic = self.last_guide.get_topic_by_title(self.recent_topic)
            if topic.has_step(step):
                if Matcher.is_valid(step):
                    line_desc = "instruction (%s)" % self.recent_topic
                    method = "yellow"
                else:
                    line_desc = "unsupported (%s)" % self.recent_topic
                    method = "red"
                args = (preview_line, "<--- %s" % line_desc)
                return getattr(Console, method)(*args)
        return Console.darkgray(preview_line)
